#!/usr/bin/env python

"""TKDO
A to-do list aggregator with due dates, repeating events, and a snooze button.

This pulls tasks from Vim-VimOutliner (.otl) files, sorts and filters them,
and outputs a single list with tasks prioritized in order.

Based vaguely on concepts from PyGTD.
"""

# Copyright (C) 2008 Scott Scriven <tkdo@toykeeper.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO: ? rename to something better...  maybe 'masterlist' or 'vo-todo'?
program_name = "tkdo"
program_version = "0.6"

cfg = None

import re

datefmt = "%Y-%m-%d@%H:%M"

def main(args):
    """TKDO: todo list aggregator and manager for vimoutliner todo.otl files
    Usage: tkdo [global options] command [command options]
    Global options include:
      -h --help    Synonym for the "help" command.
    """
    import pycfg
    global cfg
    cfg = pycfg.config(program_name)
    set_defaults(cfg)
    create_config(cfg)
    # FIXME: tends to clobber the config data if loading had any errors
    # (which means losing all the config when it saves later)
    cfg.load()

    # build a list of command names to functions, including aliases
    global _cmd_map
    _cmd_map = {}
    for name in globals():
        if name.startswith('cmd_'):
            func = globals()[name]
            _cmd_map[name[4:]] = func
            aliases = getattr(func, 'aliases', [])
            for a in aliases:
                _cmd_map[a] = func

    # handle command line
    if args:
        for i, a in enumerate(args):
            if a[0] != '-': break
            elif a in ['-h', '--help']:
                return cmd_help(*args[i+1:])
            else:
                return cmd_help()
        args = args[i:]

    if not args:
        args = [cfg.default_mode]

    op = args[0]
    try:
        func = _cmd_map[op]
    except:
        print 'Invalid command: %s' % (op)
        return

    return func(*args[1:])

def set_defaults(cfg):
    cfg.Taskfile = Taskfile

    cfg.doc(action_log="File to store actions in.")
    cfg.default(action_log="~/.%s/action_log" % (program_name))

    cfg.doc(active_bonus=["Score multiplier for 'active' tasks.",
        "(float, slightly higher than 1.0)"])
    cfg.default(active_bonus=1.25)

    cfg.doc(default_mode=["Which command to run, if none is specified on command line?",
        "(run 'tkdo help' to list possible commands)"])
    cfg.default(default_mode='gui')

    cfg.doc(due_bonus=["Score multiplier for 'due' and 'fixed recurring' tasks.",
        "(float, slightly higher than 1.0)"])
    cfg.default(due_bonus=1.5)

    cfg.doc(editor=["Command to run for external editor.",
        "Use $(FILE) where you want the file name and $(LINENO) for line number."])
    cfg.default(editor="gvim -f +$(LINENO) $(FILE)")

    cfg.doc(files="A list of Taskfile objects.")
    cfg.default(files=[])

    cfg.doc(hide_zero="Should we hide zero-score items?")
    cfg.default(hide_zero=True)

    cfg.doc(hide_done="Should we hide completed tasks?")
    cfg.default(hide_done=True)

    cfg.doc(keybindings=["Map of key to action bindings in the task list.",
        "Is a dict, with keyboard key names on the left side, ",
        "action names on the right side."])
    cfg.default(keybindings = {
        'g': 'gvim task',
        'slash': 'text search',
        # "L" is what I use in mutt
        'L': 'text search',
        'at': 'context search',
        'Escape': 'cancel search',
        'space': 'toggle expand',
        'x': 'toggle complete',
        'z': 'snooze',
        'colon': 'vi mode',
        'a': 'toggle active',
        't': 'set type',
        'i': 'set importance',
        'l': 'set lead time',
        'e': 'edit title',
        'd': 'set due date',
        'o': 'add task',
        'O': 'insert task',
        'Delete': 'delete task',
        #'k': 'skip',
        })

    cfg.doc(make_backups=["Copy task files to .bak before saving?",
        "If True, will copy file.otl to file.otl.bak before saving changes."])
    cfg.default(make_backups=True)

    cfg.doc(default_importance="Default task importance (value for I).")
    cfg.default(default_importance=50)

    cfg.doc(max_importance="Maximum task importance (value for I).")
    cfg.default(max_importance=100)

    cfg.doc(max_overdue=["Maximum score multiplier for overdue non-recurring tasks.",
        "(float, higher than 1.0)"])
    cfg.default(max_overdue=1.75)

    cfg.doc(max_overdue_fixed=["Maximum score multiplier for overdue 'fixed' recurring tasks.",
        "(float, higher than 1.0)"])
    cfg.default(max_overdue_fixed=1.75)

    cfg.doc(max_overdue_float=["Maximum score multiplier for overdue 'floating' recurring tasks.",
        "(float, higher than 1.0)"])
    cfg.default(max_overdue_float=1.333)

    cfg.doc(score_colors=["A list of (threshold, color, label) for the score column.",
        "Color can be None, to use 'default color'.",
        "Label can be None, to display the actual score.",
        "Example:",
        "score_colors=(",
        "    (100, 'red', 'NOW'),",
        "    (51, 'green', 'now'),",
        "    (5, 'yellow', 'later'),",
        "    (0, 'lightgrey', 'snoozed'),",
        "    (-1, 'darkgrey', 'done'),",
        "    )",
        ])
    cfg.default(score_colors=(
        (100, '#ff0000', None),
        (92, '#ee4400', None),
        (87, '#ffaa00', None),
        (82, '#ffcc00', None),
        (75, '#ffff00', None),
        (69, '#ccff00', None),
        (62, '#aaff00', None),
        (57, '#77ff00', None),
        (51, '#44ff00', None),
        (50, '#00ff00', None),
        (37, '#66dd66', None),
        (25, '#99cc99', None),
        (12, '#aaccaa', None),
        (5, '#bbccbb', None),
        (0, 'lightgrey', 'snooze'),
        (-1, 'darkgrey', 'done'),
        ))

    cfg.doc(show_keypresses=["Display each keypress in GUI?",
        "(can be useful when configuring key bindings)"])
    cfg.default(show_keypresses=False)

    cfg.doc(toolbar_buttons=["A list of buttons to display on the toolbar.",
        "Each item is ('label', 'action').",
        "Actions are the same as for key bindings."])
    cfg.default(toolbar_buttons=[
        ('X', 'toggle complete'),
        ('Z', 'snooze'),
        ('<--', 'toggle active'),
        ('due', 'set due_date'),
        ('i', 'set importance'),
        ('vi', 'gvim task'),
        ('@', 'show contexts'),
        ('/', 'text search'),
        ])

    cfg.doc(toolbar_button_padding="How many pixels to draw around button labels.")
    cfg.default(toolbar_button_padding=10)

    cfg.doc(toolbar_visible="Whether to show the tool bar.")
    cfg.default(toolbar_visible=False)

    cfg.doc(window_size=["Default window width and height in pixels.",
        "Should be a (x,y) tuple, such as: (640,480)"])
    cfg.default(window_size=(500,600))

    return cfg

def create_config(cfg):
    import os
    dirname = "%s/.%s" % (os.environ["HOME"], program_name)
    rcpath = "%s/rc" % (dirname)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    if not os.path.exists(rcpath):
        open(rcpath, "w")
        cfg.save()

def log_action(text):
    """Add "text" to the action log.
    (keep a log of actions, for later reporting...
    to answer 'what did I finish this week?', for example)
    """

    if not cfg.action_log:
        return
    import os
    path = os.path.expanduser(cfg.action_log)

    import time
    dts = time.strftime("%Y-%m-%d %H:%M:%S")
    line = "%s %s\n" % (dts, text)

    fp = open(path, "a+b")
    fp.write(line)
    fp.flush()
    fp.close()

def aliases(*names):
    "Assigns func.aliases = names"
    def decorate(func):
        func.aliases = names
        return func
    return decorate

def cmd_cal(*args):
    """Print a list of tasks with a due date.
    Usage: cal [options] [search terms] [@contexts]
    Options include:
      -r  --reverse   Print in reverse order (least-to-most important).
      -N              Display only the first N items.
    """
    reverse = False
    limit = None
    contexts = []
    filters = []
    for a in args:
        if a in ['-r', '--reverse']:
            reverse = True
        elif a.startswith('-'):
            try:
                limit = int(a[1:])
            except:
                return cmd_help('cal')
        elif a.startswith('@'):
            contexts.append(a[1:])
        else:
            filters.append(a)

    #cfg.hide_done = False
    cfg.hide_zero = False

    ml = Masterlist(cfg.files)
    ml.load()
    ml.sort()
    tasks = ml.filter(filters=filters, contexts=contexts)
    tasks = [t for t in tasks if (t.inherit('recur_type') in ('due', 'fixed'))]
    # sort by due date?
    tmp = [(t.inherit('due_date'),t.lineno,t) for t in tasks]
    tmp.sort()
    tasks = [t[-1] for t in tmp]

    import time
    now = time.time()
    prev_time = 0.0

    lines = []
    # calculate / show due dates
    for t in tasks:
        due = ""
        due_date = t.inherit('due_date')
        if due_date and due_date[0]:
            due = fmt_date(due_date[0])
            cur_time = time.mktime(due_date[0])
        # display a separator for the current date
        if (prev_time < now) and (cur_time >= now):
            lines.append(('-' * 16) + ' now ' + ('-' * 16))
        prev_time = cur_time
        text = "%-16s %s" % (due, t.text_oneliner())
        lines.append(text)

    if limit:
        lines = lines[:limit]

    if reverse:
        lines.reverse()

    for l in lines:
        print l

def cmd_cfg(*args):
    """Show the current configuration.
    Usage: cfg [options]
    Options include:
      -o PATH  --outpath   Write config data to PATH.
    Suggested use:
      tkdo cfg -o ~/.tkdo/rc
      vim ~/.tkdo/rc
    """
    outpath = None
    i = 0
    while i < len(args):
        a = args[i]
        if a in ['-o', '--outpath']:
            i += 1
            outpath = args[i]
        else:
            return cmd_help('cfg')
        i += 1

    if outpath:
        text = str(cfg)
        fp = open(outpath, 'wb')
        fp.write(text)
        fp.close()
    else:
        print cfg

@aliases('cx')
def cmd_contexts(*args):
    """Print a list of known contexts.
    Usage: contexts [options]
    Options include:
    """
    contexts = {}
    for a in args:
        return cmd_help('contexts')

    ml = Masterlist(cfg.files)
    ml.load()
    contexts = ml.by_context()
    names = contexts.keys()
    names.sort()
    for c in names:
        # TODO: align columns
        # TODO: show 'visible' vs total tasks
        print '@%s\t%s tasks' % (c, len(contexts[c]))

@aliases('g')
def cmd_gui(*args):
    """Run the graphical user interface.
    Usage: gui
    (has no parameters)
    """
    #ml = Masterlist(cfg.files)
    ml = MasterlistGUI(cfg.files)
    ml.load()
    ml.sort()
    #ml.print_list()
    ml.run()

@aliases('h')
def cmd_help(*args):
    """Display help.  (for example, "tkdo help gui")
    Usage: help [topic1 [topic2 ...]]
    If no topic is requested, an overview and list of commands is shown.
    """
    def getdoc(name):
        doc = ""
        func = _cmd_map[name]
        try: doc = func.__doc__
        except: pass
        if not doc: doc = "[undocumented]"
        aliases = [func.__name__[4:]]
        aliases.extend(getattr(func, 'aliases', []))
        doc = doc + "\nAliases: " + ', '.join(aliases)
        return doc

    if args:
        for a in args:
            try:
                doc = getdoc(a)
                print doc
            except:
                print "No such function: %s" % (a)
    else:
        print main.__doc__
        print "Commands include:"
        fmt = "    %-10s %s"
        cmds = [k for k in globals() if k.startswith('cmd_')]
        cmds.sort()
        for c in cmds:
            name = c.replace('cmd_', '')
            text = getdoc(name).split('\n')[0].strip()
            print fmt % (name, text)

def cmd_ical(*args):
    """Export due tasks in iCalendar format.
    Usage: ical [options] [search terms] [@contexts]
    Options include:
      -o PATH  --outfile   Write output to PATH instead of stdout.
    """
    outpath = None
    contexts = []
    filters = []

    i = 0
    while i < len(args):
        a = args[i]
        if a in ['-o', '--outfile']:
            i += 1
            outpath = args[i]
        # TODO: add an option to include full title or only leaf title
        elif a.startswith('-'):
            return cmd_help('ical')
        elif a.startswith('@'):
            contexts.append(a[1:])
        else:
            filters.append(a)
        i += 1

    cfg.hide_zero = False

    ml = Masterlist(cfg.files)
    ml.load()
    ml.sort()
    tasks = ml.filter(filters=filters, contexts=contexts)
    tasks = [t for t in tasks if (t.inherit('recur_type') in ('due', 'fixed'))]
    # sort by due date
    tmp = [(t.inherit('due_date'),t.lineno,t) for t in tasks]
    tmp.sort()
    tasks = [t[-1] for t in tmp]

    def make_uuid(t):
        date = date2str(t.due_date[0])
        lineno = t.lineno
        host = 'tkdo.example.com'
        text = 'UID:%s-%s@%s' % (date, lineno, host)
        return text

    def make_rrule(due, num, type_):
        "create recur rules"
        num = int(num)
        types = dict(h='HOURLY', d='DAILY', w='WEEKLY', m='MONTHLY', y='YEARLY')
        parts = ['FREQ=%s' % (types[type_])]
        parts.append('INTERVAL=%s' % (num))
        text = 'RRULE:%s' % (';'.join(parts))
        return text

    def date2str(date):
        "for tasks due at 00:00 or 23:59, show only date and no DTEND"
        fmt = '%Y%m%dT%H%M00'
        text = fmt_date(date, fmt)
        # chop time off all-day or any-time-of-day events
        if text[-7:] in ('T000000', 'T235900'):
            text = text[:-7]
        return text

    lines = [
            'BEGIN:VCALENDAR',
            'VERSION:2.0',
            'PRODID:-//TKDO//NONSGML TKDO v%s//EN' % (program_version),
            '',
            ]
    # calculate / show due dates
    # TODO: figure out an appropriate duration for VEVENTs (currently 1 hour)
    # TODO: use python-vobject lib instead?
    # TODO: export importance value as a priority (1-9)?
    for t in tasks:
        due_date = t.inherit('due_date')
        lines.append('BEGIN:VEVENT')
        lines.append(make_uuid(t))
        lines.append('SUMMARY:%s' % (t.full_title()))
        # export start/end times
        tstamp = date2str(due_date[0])
        if 'T' in tstamp: # specific time
            end = date2str(date_add(due_date[0], 1, 'h'))
            lines.append('DTSTART:%s' % (tstamp))
            lines.append('DTEND:%s' % (end))
        else: # no specific time
            lines.append('DTSTART:%s' % (tstamp))
        # export recur data
        if due_date[1]: # recurring task
            lines.append(make_rrule(*due_date))
        # export contexts as categories
        contexts = t.inherit_list('contexts')[:]
        contexts.insert(0, t.get_root().title) # task file title
        if contexts:
            cattext = ','.join([x.upper() for x in contexts])
            lines.append('CATEGORIES:%s' % (cattext))
        # export text details as a description
        dlines = [l for l in t.lines[1:] if not l.strip().startswith('; TKDO: ')]
        if dlines:
            desc = '\n'.join(dlines).strip()
            lines.append('DESCRIPTION:%s' % (desc))
        # done
        lines.append('END:VEVENT')
        lines.append('')

    lines.append('END:VCALENDAR')

    if outpath:
        outfile = open(outpath, "w")
    else:
        import sys
        outfile = sys.stdout

    outfile.write('\n'.join(lines))
    outfile.write('\n')

@aliases('l', 'ls')
def cmd_list(*args):
    """Print the master list in plain text format.
    Usage: list [options] [search terms] [@contexts]
    Options include:
      -N              Only show the top N items.
      -r  --reverse   Print in reverse order (least-to-most important).
      -a  --all       Show all (synonym for "-z -d")
      -z  --zero      Show zero-score (snoozed) items.
      -Z  --no-zero   Hide zero-score (snoozed) items.
      -d  --done      Show completed items.
      -D  --no-done   Hide completed items.
    """
    reverse = False
    limit = None
    contexts = []
    filters = []
    for a in args:
        if a in ['-r', '--reverse']:
            reverse = True
        elif a in ['-a', '--all']:
            cfg.hide_zero = False
            cfg.hide_done = False
        elif a in ['-z', '--zero']:
            cfg.hide_zero = False
        elif a in ['-Z', '--no-zero']:
            cfg.hide_zero = True
        elif a in ['-d', '--done']:
            cfg.hide_done = False
        elif a in ['-D', '--no-done']:
            cfg.hide_done = True
        elif a.startswith('-'):
            try:
                limit = int(a[1:])
            except:
                return cmd_help('list')
        elif a.startswith('@'):
            contexts.append(a[1:])
        else:
            filters.append(a)

    ml = Masterlist(cfg.files)
    ml.load()
    ml.sort()
    tasks = ml.filter(filters=filters, contexts=contexts)
    lines = ml.leaves_as_text(tasks)
    if limit:
        lines = lines[:limit]

    if reverse:
        lines.reverse()

    for l in lines:
        print l

def cmd_tick(*args):
    """Recalculate all task statuses and save if anything changed.
    Run this in your crontab to wake up tasks automatically.
    Usage: tick [options]
    Options include:
      -v  --verbose   Print a list of changed files.
          --dry-run   Don't actually save any changes.
      -d  --diff      Also print a unified diff of changes for each file.
    """
    verbose = False
    dryrun = False
    diffs = False
    for a in args:
        if a in ['-v', '--verbose']:
            verbose = True
        elif a in ['--dry-run']:
            dryrun = True
        elif a in ['-d', '--diff']:
            diffs = True
        else:
            return cmd_help('tick')

    ml = Masterlist(cfg.files)
    ml.load()
    ml.sort() # recalculate task data
    for tf in ml.taskfiles:
        # FIXME (maybe): shouldn't need to re-read the file
        intext = tf.raw_text()
        outtext = tf.to_otl()
        if (not outtext) or (not intext):
            print "Skipped %s: file empty or unreadable" % (tf.url)
            continue
        if intext != outtext:
            if verbose:
                print "%s changed (%s)" % (tf.title, tf.url)
            if diffs:
                show_diff(intext, outtext, name=tf.title)
            if not dryrun:
                tf.save()

def cmd_version(*args):
    """Print the program name and version.
    Usage: version
    """
    print "%s %s" % (program_name, program_version)

class Task:
    # removed 'project' concept, since I didn't use it
    valid_types = ('task', 'next', 'idea', 'ignored')
    active_types = ('task', 'next', 'idea')

    # only for recurring events, not plain old 'due' events
    recur_types = ('float', 'fixed')

    meta_pats = dict(
            type = re.compile(r"T=(\S+)"),
            # inferred from due_date
            # TODO: make HH:MM optional
            due_date = re.compile(r"D=(\d{4}-\d\d-\d\d@\d\d:\d\d)?(\+([\.\d]+)([hdwmy]))?"),
            lead_time = re.compile(r"L=([\.\d]+)([hHdDwWmMyY])"),
            snooze_time = re.compile(r"Z=(\d{4}-\d\d-\d\d@\d\d:\d\d)\+([\.\d]+)([hdwmy])"),
            last_done = re.compile(r"X=(\d)?,?(\d{4}-\d\d-\d\d@\d\d:\d\d)"),
            importance = re.compile(r"I=([\d]+)"),
            depth = re.compile(r"depth=([\d]+)"),
            )

    px_pat = re.compile(r" (\d{0,3})%($|\s)") # match completion percent
    at_pat = re.compile(r" @(\S+)") # match contexts

    def __init__(self, parent=None, lines=[], lineno=0):
        self.parent = parent
        self.lines = lines
        self.lineno = lineno

        self._reset()

        self._parse_meta()

        # create child tasks, if we're not too deep already
        depth = self.inherit('depth')
        if (self.type in self.active_types) and \
                (depth > self.level + 1): # depth is 1-based, level is 0-based
            self.load_children()

    def _reset(self):
        self.level = 0
        if self.parent:
            self.level = self.parent.level + 1

        self.tasks = []

        self.alias = None
        self.title = None
        self.active = None
        self.type = None
        self.done = None
        self.completion = None
        self.last_done = None
        self.was_done = None
        self.due_date = None
        self.recur_type = None
        self.lead_time = None
        self.snoozed = None
        self.contexts = []
        self.importance = None
        self.depth = None
        self._score = None

    def get_root(self):
        "find and return our root node"
        t = self
        while t.parent:
            t = t.parent
        return t

    def get_relation(self, rel):
        t = self
        while rel:
            if rel.startswith('t'): # this task
                rel = rel[1:]
            elif rel.startswith('p'): # parent
                rel = rel[1:]
                if t.parent:
                    t = t.parent
            elif rel.startswith('f'): # entire file
                rel = rel[1:]
                t = t.get_root()
        return t

    def load_children(self):
        indent = '\t' * (self.level + 1)
        # find next-level items
        # must be indented enough, but not too much
        # must not be our meta line
        firstlines = [num for num,line in enumerate(self.lines)
                if (line.startswith(indent))
                and (not line[len(indent):].startswith('\t'))
                and (not line[len(indent):].startswith('; TKDO: '))]

        child_tasks = []
        for tnum in range(len(firstlines)):
            start = firstlines[tnum]
            try:
                next = firstlines[tnum+1]
            except:
                next = None
            tlines = self.lines[start:next]
            t = Task(parent=self, lines=tlines, lineno=start+self.lineno)
            child_tasks.append(t)

        self.tasks = child_tasks

    def has_active_subtasks(self):
        for t in self.tasks:
            if t.type in self.active_types:
                return True
        return False

    def full_title(self):
        parts = []
        parent = self
        while parent:
            if parent.alias: text = parent.alias
            else: text = parent.title
            parts.insert(0, text)
            parent = parent.parent
        parts = [str(x) for x in parts]
        return " :: ".join(parts)

    def inherit(self, key):
        "retrieve a value from self or ancestor"
        #print "inherit(%s) %s" % (key, repr(self.full_title()))
        val = None

        parent = self
        while parent and (val is None):
            val = getattr(parent, key)
            parent = parent.parent
        #print 'inherit(): %s' % str(val)
        return val

    def inherit_list(self, key):
        "return a merged list from all ancestors"
        val = getattr(self, key)

        # collect values
        parent = self.parent
        while parent:
            pval = getattr(parent, key)
            val.extend(pval)
            parent = parent.parent

        # remove duplicates
        dv = dict(zip(val, [1]*len(val)))
        val = dv.keys()

        return val

    def _parse_meta(self):
        if len(self.lines) >= 2:
            tline, mline = self.lines[:2]
        elif len(self.lines) >= 1:
            tline = self.lines[0]
            mline = None
        else:
            tline = mline = None
        self._parse_firstline(tline)
        if self.type in self.active_types:
            self._parse_metaline(mline)

    def _parse_firstline(self, line):
        if not line:
            self.type = 'ignored'
            return

        # amount of indentation; depth in outline tree
        #self.level = counttabs(line)

        # percent done
        found = self.px_pat.search(line)
        if found:
            px = found.group(1)
            if not px: px = "0"
            self.completion = float(px) / 100.0

        # remove extra stuff from title
        mangle = (
                ('[_] ', ''),
                ('[X] ', ''),
                ('[Z] ', ''), # snoozed
                (' <--', ''),
                )
        self.title = line.strip()
        for search, replace in mangle:
            self.title = self.title.replace(search, replace)

        # determine task type and state
        lstrip = line.strip()
        if lstrip.startswith('[_] ') or lstrip.startswith('[Z] '):
            self.type = 'task'
            self.done = False
        elif lstrip.startswith('[X] '):
            self.type = 'task'
            self.done = True
        else:
            self.type = 'ignored'
            self.title = line
            self.done = False

        # leave None instead of False, so inheritance will work
        if " <--" in line: self.active = True

        # also, load (and remove) contexts
        parts = self.title.split()
        self.contexts = [p[1:] for p in parts if (p.startswith('@') and len(p) > 1)]
        self.title = self.at_pat.sub('', self.title)

    def _parse_metaline(self, line):
        """read the line containing all the task attributes,
        assigning any discovered values to self.
        example:
            ; TKDO: T=task D=2008-04-14@17:00+1m L=1w Z=2008-04-00@12:12+1w X=1,2008-03-14@17:33 @work @online
        """
        if not line:
            return

        ## ensure line is indented once more than self
        ## (turns out it's better to let the caller ensure this is the right line)
        #if counttabs(line) != counttabs(self.lines[0]) + 1:
        #    return
        if not line.strip().startswith('; TKDO: '):
            return

        #print line

        result = {}

        # contexts
        parts = line.split()
        contexts = [x[1:] for x in parts if (x.startswith('@') and len(x) > 1)]
        result['contexts'] = contexts
        # merge instead of replacing
        old = dict(zip(self.contexts, [None]*len(self.contexts)))
        new = dict(zip(contexts, [None]*len(contexts)))
        new.update(old)
        self.contexts = new.keys()

        # grab what I can via regexes
        for key, pattern in self.meta_pats.iteritems():
            found = pattern.search(line)
            if found:
                value = found.groups()
                result[key] = value

        #for key, values in result.iteritems():
        #    print "%s: %s" % (key, values)

        if 'depth' in result:
            x = result['depth'][0]
            self.depth = int(x)

        if 'type' in result:
            t = result['type'][0]
            if t in self.valid_types:
                self.type = t
            if self.type == 'idea':
                self.importance = self.inherit('importance') / 2.0

        if 'importance' in result:
            x = result['importance'][0]
            x = float(x) / 100.0
            if self.type == 'idea':
                x /= 2
            self.importance = x

        if 'last_done' in result:
            x = result['last_done']
            self.was_done = self.done
            if x[0] is not None:
                self.was_done = not not int(x[0])
            self.last_done = parse_date(x[1])

            # user checked off this item since last time we looked
            #if self.done and not self.was_done:
            #    self.last_done = time.localtime()

        if 'lead_time' in result:
            r = result['lead_time']
            self.lead_time = (float(r[0]), r[1].lower())

        if 'snooze_time' in result:
            x = result['snooze_time']
            base = parse_date(x[0])
            self.snoozed = (base, float(x[1]), x[2].lower())

        if 'due_date' in result:
            r = result['due_date']
            r = (r[0], r[2], r[3])
            base = None
            self.recur_type = None
            if r[0]:
                base = parse_date(r[0])
                self.recur_type = 'due'
                if r[1]:
                    self.recur_type = 'fixed'
            if r[1]:
                self.due_date = (base, float(r[1]), r[2].lower())
                if not base:
                    self.recur_type = 'float'
            else:
                self.due_date = (base, None, None)

        """
        if self.recur_type == 'fixed':
            # TODO: magic with self.done and adjusting our due date
            # if not done:
            #   no change
            # elif done:
            #   while last completed > due date - lead time:
            #     bump due date forward one unit
            #   if last completed < due date - lead time:
            #     no change
            #   else:
            #     done = 0
        """

    def get_completion(self):
        kids = [t for t in self.tasks if t.type in self.active_types]
        if not kids:
            if self.completion is None:
                self.completion = float(self.done)
            return self.completion

        complete = sum([t.get_completion() for t in kids])
        self.completion = complete / len(kids)
        return self.completion

    def complete(self, done=True):
        import time
        self.done = done
        self.completion = 1.0
        # FIXME: notify parent
        # FIXME: complete children

        if not self.done:
            self.completion = 0.0
            return

        self.last_done = time.localtime()

        # update our due date
        if self.recur_type in self.recur_types:
            due_date = self.due_date
            last_done = time.mktime(self.last_done)
            lead = 0.0 # TODO: unsure if this is correct
            interval = due_date[1:3]
            if self.recur_type == 'float':
                # no explicit due date
                pass
            elif self.recur_type == 'fixed':
                due = due_date[0]
                while (time.mktime(due) - lead) < last_done:
                    due = date_add(due, *interval)

                foo = list(due_date)
                foo[0] = due
                self.due_date = tuple(foo)

    def snooze(self, howlong):
        """Mark this item as snoozed, starting now, for 'howlong' time.
        'howlong' should be a number and a time unit [hdwmy]
        """
        pat = re.compile(r"^\+?([\.\d]*)\s*([hHdDwWmMyY])$")

        found = pat.search(howlong)
        if not found:
            raise ValueError, "Invalid snooze length."

        num = found.group(1)
        if num:
            num = float(num)
        else:
            num = 1.0
        unit = found.group(2).lower()

        import time
        now = time.localtime()
        self.snoozed = (now, num, unit)

    def set_active(self, state=None):
        if state is None:
            # toggle
            if self.active:
                self.active = None
            else:
                self.active = True
        else: # set to True or None (not False)
            if state:
                self.active = True
            else:
                self.active = None

    def set_type(self, type_=None):
        if type_ is None:
            type_ = 'task'
        if type_ not in self.active_types:
            raise ValueError, "unknown task type"

        old = self.type
        self.type = type_

        # handle transitions
        if type_ == 'idea':
            if old != type_:
                self.importance = self.inherit('importance') / 2.0
        else:
            if (old == 'idea') and (type_ != old):
                self.importance = self.inherit('importance') * 2.0

    def set_importance(self, imp=None):
        if imp is None: self.importance = None
        else:
            imp = min(cfg.max_importance, max(0,imp))
            self.importance = imp / 100.0
        return int(self.inherit('importance') * 100)

    def add_sibling(self, title, after=True, copyme=True):
        lines = [str(title)]
        new_task = Task(self.parent, lines)

        # copy my attributes to new task
        if copyme:
            for attr in ("active", "alias", "contexts", "depth", "due_date", 
                    "importance", "lead_time", "recur_type", "type"):
                setattr(new_task, attr, getattr(self, attr))

        if not self.parent: # if we're a task file
            if after:
                self.tasks.append(new_task)
            else:
                self.tasks.insert(0,new_task)
            return

        # find our place in self.parent
        found = False
        for i,t in enumerate(self.parent.tasks):
            if t is self:
                found = True
                break
        if not found:
            i=0
        if after:
            i += 1
        self.parent.tasks.insert(i, new_task)

        return new_task

    def make_metaline(self):
        "generate a meta line"
        data = []

        # somewhat ignores same-as-parent attributes while saving

        # type
        if self.type != 'task':
            data.append('T=%s' % (self.type))

        # depth
        if self.depth is not None:
            data.append('depth=%s' % (self.depth))

        # importance
        # (only save if non-default, and different than parent)
        imp = self.inherit('importance')
        pimp = cfg.default_importance / 100.0
        if self.parent:
            pimp = self.parent.inherit('importance')
        if self.type == 'idea': imp *= 2
        if (self.importance is not None)  \
                and  (pimp != imp) \
                and  ((imp < pimp-0.001) or (imp > pimp+0.001)):
            data.append("I=%s" % (int(imp*100)))

        # due date / recurrance info
        if self.due_date:
            x = self.due_date
            if x[0] and x[1]:
                base = fmt_date(x[0])
                data.append("D=%s+%s%s" % (base, iround(x[1]), x[2]))
            elif x[0]:
                base = fmt_date(x[0])
                data.append("D=%s" % (base))
            elif x[1]:
                data.append("D=+%s%s" % (iround(x[1]), x[2]))
            else:
                # this shouldn't happen
                self.due_date = None

        # lead time
        if self.lead_time:
            x = self.lead_time
            data.append("L=%s%s" % (iround(x[0]), x[1]))

        # last completion date
        if self.done and not self.last_done:
            data.append('X=%s,%s' % (int(self.done),fmt_date()))
        elif self.last_done:
            data.append('X=%s,%s' % (int(self.done),fmt_date(self.last_done)))

        # snooze data
        if self.snoozed:
            x = self.snoozed
            # omit, if snooze end time has passed
            end = date_add(*x)
            import time
            if end > time.localtime():
                base = fmt_date(x[0])
                data.append("Z=%s+%s%s" % (base, iround(x[1]), x[2]))

        # contexts
        pcontexts = []
        if self.parent: pcontexts = self.parent.inherit_list('contexts')
        done = []
        contexts_sorted = self.contexts[:]
        contexts_sorted.sort()
        for c in contexts_sorted:
            if (c not in pcontexts) and (c not in done):
                data.append('@%s' % (c))
                done.append(c)

        line = ('\t' * (self.level+1)) + '; TKDO: ' + " ".join(data)
        # avoid meta lines with nothing but an X tag
        # (those just don't matter)
        if (len(data) == 1) and (data[0].startswith('X=')):
            return None

        elif data:
            return line

        return None

    def to_otl(self):
        """Return text representing self in vimoutliner .otl format.
        """
        # ignore same-as-parent attributes while saving
        if self.type == 'deleted':
            return ''
        if self.type == 'ignored':
            text = "\n".join(self.lines)
            return text

        import time

        lines = self.lines[:]
        depth = self.level
        # build a main line
        line = self.title

        # update 'done' based on completion
        complete = self.get_completion()
        if complete > 0.99999:
            self.done = True
        else:
            self.done = False

        # choose a task marker
        active = self.active
        if self.type in self.active_types:
            # TODO: clean up this messy logic
            if self.is_snoozed():
                if self.done and (self.recur_type not in self.recur_types):
                    line = '[X] ' + line
                else:
                    line = '[Z] ' + line
            elif self.recur_type:
                if self._score is None: self.score()
                if self.done and (self.recur_type == 'due'):
                    line = '[X] ' + line
                elif self.done or (self._score < 0.001):
                    line = '[Z] ' + line
                else:
                    line = '[_] ' + line
            elif self.done:
                line = '[X] ' + line
                active = False # completed one-time tasks are never active
            else:
                line = '[_] ' + line

        # update our percent done
        found = self.px_pat.search(line)
        if found:
            x = r" %s%%\2" % (int(100*complete))
            line = self.px_pat.sub(x, line)

        # show active tasks
        if active: line = line + " <--"

        # finish up
        line = ('\t' * depth) + line
        lines[0] = line

        # add a meta line
        line = self.make_metaline()
        if line:
            lines.insert(1, line)
            lead = lines[:2]
            remaining = lines[2:]
        else:
            lead = lines[:1]
            remaining = lines[1:]

        # recurse into child tasks when depth allows
        if self.tasks:
            lines = lead
            for t in self.tasks:
                if t.type == 'deleted': continue
                lines.append(t.to_otl())

        else:
            # FIXME: extra meta lines aren't always removed
            # (and this is a klunky approach anyway)
            # remove old/extra meta lines
            lines = lead + [l for l in remaining
                    if not l.strip().startswith('; TKDO: ')
                    or (counttabs(l) != depth+1) ]

        return "\n".join(lines)

    def text_oneliner(self):
        active = ""
        if self.inherit('active'): active = " <--"
        done = ""
        if self.done: done = "[X] "
        if self._score is None:
            self.score()
        text = "%3s %s%s%s" % (max(0, int(self._score * 100)), done, self.full_title(), 
                active)
        return text

    def is_snoozed(self):
        import time
        now = time.time()

        snoozed = self.inherit('snoozed')
        if snoozed:
            snoozed_at = snoozed[0]
            interval = snoozed[1:3]
            due = date_add(snoozed_at, *interval)
            due = time.mktime(due)
            waketime = due - (duration(*interval) / 2.0)
            if waketime > now: # not time to wake up
                return True

            if now > due: # snooze has passed entirely
                self.snoozed = None

        return False

    def timeliness(self, when=None):
        """Compute the urgency, or scheduling pressure for this task.
        Returns a float from 0.0 to a maximum urgency (normally 1.5).
        """
        # FIXME: should probably eliminate side effects
        import time
        t = 1.0
        if when:
            now = when
        else:
            now = time.time()

        # snooze hides an item for snooze_length / 2,
        # then lets it bubble up for the other half of the snooze_length.
        parent = self
        while parent: # snooze from all parents, not just the first match
            snoozed = parent.snoozed
            if snoozed:
                start = time.mktime(snoozed[0])
                # due = start + length
                due = date_add(snoozed[0], *snoozed[1:3])
                due = time.mktime(due)
                length = due - start
                window = length/2
                until = due - now
                urgency = 1.0 - (until / window)
                urgency = min(1.0, max(0.0, urgency)) # clip to range (0, 1)
                t *= urgency

            parent = parent.parent

        # due or recurring tasks
        due_date = self.inherit('due_date')
        if due_date is not None:
            interval = due_date[1:3]

            # figure out how long our ramp-up window is
            lead_time = self.inherit('lead_time')
            if lead_time:
                lead = duration(*lead_time)
            elif interval[1]:
                lead = duration(*interval) / 2.0
            else:
                lead = duration(1,'m') # default 1 month lead time

            # figure out when we were last completed
            last_done = self.inherit('last_done')
            done = self.inherit('done')
            if last_done: # has history
                last_done = time.mktime(last_done)
            elif done: # completed, but no history attached
                # assume it was completed just now
                # FIXME: write to parent here instead?
                self.last_done = time.localtime(now)
                last_done = time.mktime(self.last_done)
            else: # has never been completed
                last_done = 0.0

            # TODO: clean up this messy due date calculation code

            recur_type = self.inherit('recur_type')
            # handle plain-old-due events
            if recur_type == 'due':
                due = due_date[0]
                due = time.mktime(due)
                start = due - lead

                if start > now: # hide if it hasn't reached the lead-in time
                    t = 0.0
                else: # start < now < due :
                    until = due - now
                    urgency = 1.0 - (until / lead)
                    urgency *= cfg.due_bonus
                    urgency = min(cfg.max_overdue, max(0.0, urgency)) # clip to range (0, max_overdue)
                    t *= urgency

            # handle floating recurring events
            elif recur_type == 'float':
                due = last_done + duration(*interval)
                start = due - lead
                if start > now: # hide if it hasn't reached the lead-in time
                    t = 0.0
                else: # (start < now < due) or (past due)
                    until = due - now
                    urgency = 1.0 - (until / lead)
                    urgency = min(cfg.max_overdue_float, max(0.0, urgency)) # clip to range (0, max_overdue_float)
                    t *= urgency
                    #print "urgency: %s" % (urgency)

            # handle fixed recurring events
            elif recur_type == 'fixed':
                # base = due date = original recur date
                # while due date < date last completed:
                #     due date = due date + recur interval
                # ? recur date = due date
                due = due_date[0]
                while (time.mktime(due) - lead) < last_done:
                    due = date_add(due, *interval)

                foo = list(due_date)
                foo[0] = due
                # FIXME: write to parent here instead?
                self.due_date = tuple(foo)

                due = time.mktime(due)
                start = due - lead

                if start > now: # hide if it hasn't reached the lead-in time
                    t = 0.0
                else: # start < now < due :
                    until = due - now
                    urgency = 1.0 - (until / lead)
                    urgency *= cfg.due_bonus
                    urgency = min(cfg.max_overdue_fixed, max(0.0, urgency)) # clip to range (0, max_overdue_fixed)
                    t *= urgency
                    #print "urgency: %s" % (urgency)

        #print "timeliness: %s" % (t)
        return t

    def score(self, when=None):
        # plain old tasks with no special attributes are omitted if completed
        if self.done and (self.type in self.active_types) \
                and (self.recur_type not in self.recur_types):
                    self._score = -1.0
                    return self._score
        else:
            self._score = self.inherit('importance') * self.timeliness(when)
        if self.inherit('active'):
            self._score *= cfg.active_bonus
        return self._score

    def delete(self):
        """Prevent this task from being saved."""
        self.type = 'deleted'

class Taskfile(Task):
    def __init__(self, title, url, **kwargs):
        self.title = title
        self.url = url
        Task.__init__(self)
        self._reset()

    def _reset(self):
        tmp = self.title, self.url
        Task._reset(self)
        self.title, self.url = tmp
        self.leaf_tasks = []

        self.parent = None # should always be None
        self.type = 'task'
        self.level = -1 # 0 is a non-indented task; we're one up from that
        self.depth = 1 # top-level tasks only
        self._changed = 0 # track whether we need to save
        if self.importance is None:
            self.importance = cfg.default_importance / 100.0

    def __repr__(self):
        text = "Taskfile(%s, %s)" % (repr(self.title), repr(self.url))
        return text

    def to_otl(self):
        text = []
        # delete our meta line
        tasks = self.tasks
        try:
            assert(tasks[0].lines[0].startswith('; TKDO: '))
            del tasks[0]
        except:
            pass

        # make a new meta line
        # TODO: put metaline at end of file
        line = self.make_metaline()
        if line:
            text.append(line.strip())

        # add the rest
        for t in tasks:
            if t.type == 'deleted': continue
            text.append(t.to_otl())

        # remove trailing blank lines
        while text[-1:] == ['']:
            del text[-1]
        # ensure file ends with '\n'
        if text  and  (not text[-1].endswith('\n')):
            text.append('')

        text = "\n".join(text)
        return text

    def raw_text(self):
        "Read this task file from disk and return its contents as a string."
        fp = open_url(self.url)
        text = fp.read()
        fp.close()
        return text

    def load(self):
        self._reset()

        text = self.raw_text()
        self.lines = text.split("\n")

        self.tasks = []
        self.leaf_tasks = []

        # parse taskfile attributes from first line of file
        # TODO: grab meta line from either beginning or end of file
        if (len(self.lines) > 0) and self.lines[0].startswith('; TKDO: '):
            line = self.lines[0]
            self._parse_metaline(line)

        if (self.type == 'ignored') or (self.depth < 1):
            # don't try to load tasks
            return

        # find top-level items
        toplines = [num for num,line in enumerate(self.lines) if (not line.startswith('\t'))]

        root_tasks = []
        for tnum in range(len(toplines)):
            start = toplines[tnum]
            try:
                next = toplines[tnum+1]
            except:
                next = None
            tlines = self.lines[start:next]
            t = Task(parent=self, lines=tlines, lineno=start+1)
            root_tasks.append(t)
        self.tasks = root_tasks

        leaves = []
        def add_leaves(mytasks, parent_next=False):
            for t in mytasks:
                next = parent_next
                if t.type == 'next':
                    next = True
                if not t.has_active_subtasks():
                    leaves.append(t)
                    # If a task is explicitly snoozed, it should be considered 
                    # the 'next' task, and hides everything after it.  But if 
                    # it simply has a 0 score because it's not due yet, it 
                    # should not count as the 'next' task.
                    if parent_next and (not t.done) \
                            and (t.snoozed \
                                or (t.timeliness() > 0.001)) \
                            and (t.type in Task.active_types):
                        # only include the first subtask
                        return 1
                else:
                    found = add_leaves(t.tasks, next)
                    if found and parent_next:
                        # stop, if any unfinished tasks were found
                        return found
        add_leaves(root_tasks)
        self.leaf_tasks = leaves

    def save(self):
        # TODO: move old file to .bak first
        # FIXME: recurring tasks should be marked done or not before saving
        text = self.to_otl()
        fp = open_url(self.url, "wb")
        fp.write(text)
        fp.close()

class Masterlist:
    def __init__(self, taskfiles=None):
        self.taskfiles = []
        if taskfiles:
            self.taskfiles = taskfiles[:]
        self.tasks = []
        self.filtered_tasks = []
        self.filters = []
        self.contexts = []

    def load(self):
        for tf in self.taskfiles:
            #print repr(tf)
            tf.load()
            #print "  %s lines" % (len(tf.lines))
        self.update()

    def update(self):
        self.tasks = []
        for tf in self.taskfiles:
            tasks = [t for t in tf.leaf_tasks if t.type in Task.active_types]
            self.tasks.extend(tasks)
            #unfinished = [t for t in tasks if not t.done]
            #finished = [t for t in tasks if t.done]
            #active = [t for t in tasks if t.active]
            #print "  %s tasks (%s/%s active, %s done)" % (len(tasks), len(active), 
            #        len(unfinished), len(finished))

            #for t in tasks:
            ##    print "  %s %s: %s" % (int(t.score() * 100), t.parent.title, t.title)
            #    print t.to_otl()

    def sort(self):
        # freeze time while scoring...
        # otherwise, same-score scheduled items end up in reverse order
        # (because time passed between samples)
        import time
        now = time.time()
        tasks = [(t.score(now), -t.lineno, t.title, t) for t in self.tasks]
        tasks.sort()
        tasks.reverse()
        tasks = [t[-1] for t in tasks]
        self.tasks = tasks

    def filter(self, filters=[], contexts=[]):
        self.filters = filters
        self.contexts = contexts

        self.filtered_tasks = []

        # show/hide zero-score items
        # show/hide done items
        for t in self.tasks:
            if cfg.hide_done  and  t.done  and  (t._score < 0.001):
                continue
            if cfg.hide_zero  and  (not t.done)  and  (t._score < 0.001):
                continue
            #if cfg.hide_zero  and  (t._score < 0.001)  \
            #        and  (t.inherit('snoozed')  or  (t.inherit('recur_type') 
            #        in t.recur_types)):
            #    continue

            done = ""
            if t.done: done = "[X] "
            active = ""
            if t.inherit('active'): active = " <--"
            title = "%s%s%s" % (done, t.full_title(), active)

            skipped = False
            for search in self.filters: # filter by title
                if search.lower() not in title.lower():
                    skipped = True
                    break

            # filter by context
            t_contexts = " ".join([x.lower() for x in t.inherit_list('contexts')])
            for search in self.contexts:
                if search.lower() not in t_contexts:
                    skipped = True
                    break

            if not skipped:
                self.filtered_tasks.append(t)

        return self.filtered_tasks

    def by_context(self):
        """Return a dict whose keys are context names,
        and values are a list of tasks in that context.
        """
        contexts = {}
        for t in self.tasks:
            for c in t.inherit_list('contexts'):
                if c in contexts:
                    contexts[c].append(t)
                else:
                    contexts[c] = [t]
        return contexts

    def leaves_as_text(self, tasks=None):
        if tasks is None:
            tasks = self.tasks

        result = []
        for t in tasks:
            text = t.text_oneliner()
            result.append(text)
        return result

class MasterlistTUI(Masterlist):
    def __init__(self, *args, **kwargs):
        Masterlist.__init__(self, *args, **kwargs)

class MasterlistGUI(Masterlist):
    def __init__(self, *args, **kwargs):
        Masterlist.__init__(self, *args, **kwargs)
        import pygtk
        pygtk.require('2.0')
        import gtk
        self.gtk = gtk
        try:
            import hildon
            self.hildon = hildon
        except:
            self.hildon = None

    def run(self):
        self._build_ui()
        self.update_treestore()
        self.treeview.set_cursor(0)
        try:
            self.gtk.main()
        except KeyboardInterrupt:
            pass

    def update_treestore(self, *args):
        path, column = self.treeview.get_cursor()
        prev_task = None
        if path:
            prev_task = self.treestore[path][0]
        self.treestore.clear()

        cursor_placed = False

        filtered = self.filter(filters=self.filters, contexts=self.contexts)
        for t in filtered:
            done = ""
            if t.done: done = "[X] "
            active = ""
            if t.inherit('active'): active = " <--"
            title = "%s%s%s" % (done, t.full_title(), active)

            score = int(t.score()*100)
            color = None
            label = None
            for threshold, color_, label_ in cfg.score_colors:
                color = color_
                label = label_
                if score >= threshold:
                    break
            if not label: label = str(score)
            data = [t, label, title, color]
            branch = self.treestore.append(None, data)

            # show all contexts, including inherited ones
            contexts = t.inherit_list('contexts')
            contexts.sort()
            if contexts:
                text = ' '.join(['@'+str(c) for c in contexts])
                data = [t, '', text, None]
                self.treestore.append(branch, data)

            if len(t.lines) > 1: # display task details
                data = [t, '', "\n".join(dedent(t.lines[1:])), None]
                self.treestore.append(branch, data)

            if t == prev_task: # keep the cursor on the same item
                self.treeview.set_cursor(self.treestore.get_path(branch))
                cursor_placed = True

        # if we failed by ID, try to match by title
        if prev_task and (not cursor_placed):
            for i in range(len(self.treestore)):
                t = self.treestore[i][0]
                if t.full_title() == prev_task.full_title():
                    self.treeview.set_cursor(i)
                    cursor_placed = True
                    break

        # if we failed earlier, at least put the cursor on the same row number
        if (not cursor_placed) and path:
            self.treeview.set_cursor(path)
            cursor_placed = True

        if not cursor_placed: # last resort: top of the list
            self.treeview.set_cursor(0)
            cursor_placed = True

        # set window title to indicate number of visible tasks
        num_visible = len(filtered)
        num_total = len(self.tasks)
        num_done = len([t for t in self.tasks if t.done and (t.recur_type not in t.recur_types)])

        # show current filters too
        filters = self.filters[:]
        filters.extend([('@' + x) for x in self.contexts])
        ftext = ", ".join(filters)
        self.window.set_title("%s: %s/%s tasks (%s done) in %s files [%s]" % 
                (program_name, num_visible, num_total, num_done, 
                    len(self.taskfiles), ftext))

    def _build_ui(self):
        g = self.gtk

        if self.hildon:
            self.app = self.hildon.Program()
            self.window = self.hildon.Window()
            self.app.add_window(self.window)
        else:
            self.window = g.Window(g.WINDOW_TOPLEVEL)
        self.window.set_title(program_name)
        self.window.connect("destroy", self._quit)
        self.window.connect("delete_event", self._quit)
        self.window.set_size_request(220,220) # minimum size
        self.window.set_default_size(*cfg.window_size) # actual size
        self.window.connect("size-allocate", self._on_size_allocate)
        self.window.connect("window-state-event", self._on_window_state_change)

        vb = g.VBox(homogeneous=False, spacing=0)
        self.window.add(vb)

        self.toolbar = self._build_toolbar()
        vb.pack_end(self.toolbar, False, True, 0)

        menus = self._build_menus()
        if self.hildon: self.window.set_menu(menus)
        else: vb.pack_start(menus, False, True, 0)

        hb = g.HBox(homogeneous=False, spacing=0)
        vb.pack_start(hb)

        # task/divider, score, text, score color
        self.treestore = g.TreeStore(object, str, str, str)
        self.treeview = g.TreeView(self.treestore)
        self.treeview.set_enable_search(False)
        self.treeview.connect('key-press-event', self._tree_keypress)
        # FIXME: scrollbar can be invisible, off the right side
        # FIXME: scroll wheel doesn't work
        # FIXME: scrollbar use doesn't keep cursor onscreen
        hb.pack_start(self.treeview)
        va = self.treeview.get_vadjustment()
        scroll = g.VScrollbar(va)
        hb.pack_start(scroll, expand=False, fill=False)

        colnum = 0

        # just the widget to expand the branch
        self.expand_column = g.TreeViewColumn('')
        self.treeview.append_column(self.expand_column)
        colnum += 1

        # task score
        self.score_column = g.TreeViewColumn('score')
        self.treeview.append_column(self.score_column)
        cell = g.CellRendererText()
        self.score_column.pack_start(cell, True)
        self.score_column.add_attribute(cell, 'text', colnum)
        # set color for cell
        self.score_column.add_attribute(cell, 'background', 3)
        colnum += 1

        # task title
        self.title_column = g.TreeViewColumn('title')
        self.treeview.append_column(self.title_column)
        cell = g.CellRendererText()
        self.title_column.pack_start(cell, True)
        self.title_column.add_attribute(cell, 'text', colnum)
        colnum += 1

        # status bar area
        hb = g.HBox()
        vb.pack_start(hb, expand=False, fill=False)

        # status bar
        self._status = g.Label()
        self._status.set_justify(g.JUSTIFY_LEFT)
        self._status.set_alignment(0.0, 0.5)
        self._status.set_padding(5,5)
        hb.pack_start(self._status)

        # entry box for search, snooze, vim command emulation, etc
        self._status_entry = g.Entry()
        self._status_entry.connect('activate', self._activate_status_entry)
        hb.pack_start(self._status_entry)

        self.window.show_all()
        if not cfg.toolbar_visible: self.toolbar.hide()
        self._status_entry.hide()
        self.treeview.grab_focus()

    def _build_toolbar(self):
        toolbar = self.gtk.Toolbar()
        toolbar.set_style(self.gtk.TOOLBAR_TEXT)

        for text, action in cfg.toolbar_buttons:
            handler = self._lookup_action(action)
            button = self.gtk.ToolButton()
            label = self.gtk.Label(text)
            label.set_padding(cfg.toolbar_button_padding, cfg.toolbar_button_padding)
            button.set_label_widget(label)
            button.connect('clicked', handler)
            toolbar.insert(button, -1)

        return toolbar

    def _build_menus(self):
        "create some menu widgets"
        # FIXME: DeprecationWarning: use gtk.UIManager instead
        menu_items = (
            ('/_File', None, None, 0, '<Branch>'),
            ("/File/_Save current task's file", '<control>S', self._save_task_cb, 0, None),
            ('/File/Save _Config file', None, self._save_config_file, 0, None),
            ('/File/Re_fresh', None, self._refresh, 0, None),
            ('/File/_Reload All', '<control>R', self._action_reload_all, 0, None),
            ('/File/_Quit', '<control>Q', self._quit, 0, None),
            ('/_Edit', None, None, 0, '<Branch>'),
            ('/Edit/_Go to (edit) taskfile', '<control>G', self._action_gvim_task, 0, None),
            ('/_View', None, None, 0, '<Branch>'),
            #('/View/Expand or Collapse Task', None, self._action_toggle_expand, 0, None),
            ('/View/_Filter by Title', '<control>F', self._action_text_search, 0, None),
            ('/View/Filter by Context (_List)', None, self._action_show_contexts, 0, None),
            ('/View/Filter by _Context (Text)', None, self._action_context_search, 0, None),
            ('/View/Hide\/Show snoo_zed tasks', None, self._action_toggle_show_snoozed,  0, None),
            ('/View/Hide\/Show _done tasks', None, self._action_toggle_show_done, 0, None),
            ('/View/Hide\/Show toolbar', None, self._action_toggle_toolbar, 0, None),
            ('/View/Toggle Fullscreen', 'F6', self._action_toggle_fullscreen, 0, None),
            ('/_Actions', None, None, 0, '<Branch>'),
            ('/Actions/_Edit title', None, self._action_edit_title,0, None),
            ('/Actions/Snoo_ze item', '<control>Z', self._action_snooze,0, None),
            ('/Actions/Mark item as done', '<control>X', self._action_toggle_complete, 0, None),
            ('/Actions/Toggle _Active', None, self._action_toggle_active, 0, None),
            ('/Actions/Set _Type', '<control>T', self._action_set_type, 0, None),
            ('/Actions/Set _Importance', '<control>I', self._action_set_importance, 0, None),
            ('/Actions/Set _Due Date', '', self._action_set_due_date, 0, None),
            ('/Actions/Set _Lead Time', '', self._action_set_lead_time, 0, None),
            ('/Actions/Add _new task after current', '', self._action_add_task, 0, None),
            ('/Actions/Insert new task before current', '', self._action_insert_task, 0, None),
            ('/Actions/Delete task', '', self._action_delete_task, 0, None),
            # TODO: edit contexts
            # TODO: set due date (needs a date/duration chooser)
            ('/_Help', None, None, 0, '<LastBranch>'),
            ('/Help/Show _Keybindings', None, self._action_show_keybindings, 0, None),
            )
        g = self.gtk
        accel_group = g.AccelGroup()
        if self.hildon: factory = g.ItemFactory(g.Menu, '<main>', accel_group)
        else: factory = g.ItemFactory(g.MenuBar, '<main>', accel_group)
        factory.create_items(menu_items)
        self.window.add_accel_group(accel_group)
        self.menu_factory = factory
        menubar = factory.get_widget('<main>')
        return menubar

    def _tree_keypress(self, widget, event):
        keyval = event.keyval
        name = self.gtk.gdk.keyval_name(keyval)
        string = event.string
        #print "key: %s, %s, %s" % (keyval, name, string)
        if name in cfg.keybindings:
            try:
                func = self._lookup_action(cfg.keybindings[name])
                func()
            except Exception, e:
                self.set_status(str(e))
                raise

        else:
            if cfg.show_keypresses:
                self.set_status('Key pressed: %s' % (name))

    def _get_current_task(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            self.set_status('Select an item first.')
            return None
        data = self.treestore[path]
        t = data[0]
        if not t:
            self.set_status('Selected item is not a task.')
            return None
        return t

    def _on_size_allocate(self, widget, allocation):
        cfg.window_size = (allocation.width, allocation.height)

    def _on_window_state_change(self, widget, event, *args):
        if event.new_window_state & self.gtk.gdk.WINDOW_STATE_FULLSCREEN:
            self._fullscreen = True
        else:
            self._fullscreen = False

    def _save_task_cb(self, *args):
        t = self._get_current_task()
        if not t:
            return
        tf = t.get_root()
        self._save_task(t, "Saved taskfile: %s" % (tf.title))

    def _save_config_file(self, *args):
        self.set_status("Saving config...")
        try:
            cfg.save()
        except Exception, e:
            self.set_status(str(e))
            raise
        self.log_status("Saved config file.")

    def _save_task(self, t, msg):
        # find task file
        tf = t.get_root()
        # save
        self.set_status("Saving %s" % (tf.url))
        try:
            tf.save()
        except Exception, e:
            self.set_status(str(e))
            raise
        self.set_status("Reloading %s" % (tf.url))
        tf.load()
        self._refresh()
        self.log_status(msg)

    def gtk_flush(self, *args):
        while self.gtk.events_pending():
            self.gtk.main_iteration()

    def _refresh(self, *args):
        self.update()
        self.sort()
        self.update_treestore()

    def set_status(self, text):
        self._status.set_text(text)
        self.gtk_flush()

    def log_status(self, text):
        log_action(text)
        self.set_status(text)

    def _get_rel(self, t, text):
        """Parse 'text' for task parent symbols.
        Returns the appropriate task, and 'text' minus the symbols.
        symbols are: t, p, f  (for task, parent, file)
        """
        rel = ''
        # FIXME: only allow one type of letter...
        # 'ptfp' should set rel to 'p' and text to 'tfp'
        while text and (text[:1] in 'tpf'):
            rel = rel + text[:1]
            text = text[1:]
        if rel:
            t = t.get_relation(rel)
        return t, text

    def edit_taskfile(self, tf, t):
        url = tf.url
        lineno = t.lineno
        # FIXME: escape command parameters
        # FIXME: make local copy, edit, save via url abstraction
        cmd = str(cfg.editor)
        cmd = cmd.replace(r'$(LINENO)', str(lineno))
        cmd = cmd.replace(r'$(FILE)', str(url))
        import os
        self.log_status("Exec: %s" % (cmd))
        os.system(cmd)
        self.set_status("Done: " + cmd)

    def _lookup_action(self, name):
        """Return the _action_* function matching 'name'"""
        name = '_action_' + name.replace(' ', '_')
        func = getattr(self, name)
        return func

    def _action_refresh(self, *args):
        return self._refresh()

    def _action_reload_all(self, *args):
        for tf in self.taskfiles:
            self.set_status("Loading %s" % (tf.url))
            tf.load()
        self._refresh()
        self.set_status("Reloaded all task files.")

    def _action_gvim_task(self, *args):
        t = self._get_current_task()
        if not t:
            return

        tf = t.get_root()
        self.edit_taskfile(tf, t)
        tf.load()
        self._refresh()

    def _action_toggle_complete(self, *args):
        t = self._get_current_task()
        if not t:
            return

        # TODO: show a brief strike-thru
        import time

        # TODO: add UI for setting item completion percent
        if t.done: # mark NOT done
            t.complete(False)
            msg = "Unchecked"
        else: # mark done
            t.complete()
            msg = "Completed"

        # save
        self._save_task(t, "%s: %s" % (msg, t.full_title()))

    def _action_toggle_active(self, *args):
        t = self._get_current_task()
        if not t:
            return

        # toggle active
        t.set_active()

        if t.active: msg = "Set active: %s"
        else: msg = "Set inactive: %s"
        self._save_task(t, msg % (t.full_title()))

    def _action_toggle_show_snoozed(self, *args):
        cfg.hide_zero = not cfg.hide_zero
        self._refresh()

    def _action_toggle_show_done(self, *args):
        cfg.hide_done = not cfg.hide_done
        self._refresh()

    def _action_toggle_toolbar(self, *args):
        cfg.toolbar_visible = not cfg.toolbar_visible
        if cfg.toolbar_visible: self.toolbar.show()
        else: self.toolbar.hide()

    def _action_show_contexts(self, *args):
        """Show a list of contexts, with a checkbox by each.
        Filter the tasks if "OK" is pressed.
        """
        g = self.gtk

        contexts = self.by_context()
        names = contexts.keys()
        names.sort()

        checkboxes = []

        # build a table of (checkbox/name, summary) rows
        t = g.Table(1,2)
        row = 0
        for c in names:
            cb = g.CheckButton("@" + c)
            for look in self.contexts:
                if look in c: # allow partial matches
                    cb.set_active(True)
            checkboxes.append((c,cb))
            t.attach(cb, 0,1, row,row+1)
            t.attach(g.Label("%s items" % (len(contexts[c]))), 1,2, row,row+1)
            row += 1

        # TODO: make 'enter' and 'ctrl-enter' activate the OK button
        d = g.Dialog("Context List", parent=self.window, flags=g.DIALOG_MODAL,
                buttons=(g.STOCK_CANCEL, g.RESPONSE_REJECT, g.STOCK_OK, 
                    g.RESPONSE_ACCEPT))
        #d.vbox.pack_start(g.Label("Filter by (AND):"), False, False)
        d.vbox.pack_start(t, True, True)
        d.show_all()

        # wait for the dialog to complete
        response_id = d.run()

        d.hide()

        if response_id == g.RESPONSE_ACCEPT:
            self.contexts = []
            for name, cb in checkboxes:
                if cb.get_active():
                    self.contexts.append(name)
            self._refresh()

    def _action_show_keybindings(self, *args):
        """Display a list of current keybindings.
        """
        fmt = "%-20s %s"
        lines = []
        lines.append(fmt % ('Key', 'Action'))
        lines.append(fmt % ('-----', '-----'))
        names = cfg.keybindings.keys()
        names.sort()
        for key in names:
            action = cfg.keybindings[key]
            lines.append(fmt % (key, action))
        lines.append("")
        lines.append("")

        lines.append(fmt % ('Action', 'Key(s)'))
        lines.append(fmt % ('-----', '-----'))
        actions = {}
        for name in dir(self):
            if name.startswith("_action_"):
                name = name.replace("_action_", "").replace("_", " ")
                actions[name] = [key for key in cfg.keybindings if cfg.keybindings[key] == name]

        names = actions.keys()
        names.sort()
        for action in names:
            keys = ", ".join(actions[action])
            if not keys:
                keys = "(none)"
            lines.append(fmt % (action, keys))
        text = '<tt>' + '\n'.join(lines) + '</tt>'

        g = self.gtk
        d = g.Dialog("Keybindings", parent=self.window, flags=g.DIALOG_NO_SEPARATOR)
        l = g.Label()
        l.set_markup(text)
        # put the text in a scrollable area
        sw = g.ScrolledWindow()
        sw.add_with_viewport(l)
        d.vbox.pack_start(sw, True, True)
        d.set_default_size(400,300)

        d.show_all()

    def _action_toggle_expand(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            return
        if self.treeview.row_expanded(path):
            self.treeview.collapse_row(path)
        else:
            self.treeview.expand_row(path, open_all=False)

    def _action_toggle_fullscreen(self, *args):
        if self._fullscreen: self.window.unfullscreen()
        else: self.window.fullscreen()

    def _action_text_search(self, *args):
        self._start_entry('search', "Search:")

    def _action_context_search(self, *args):
        self._start_entry('context_filter', "Context:")

    def _action_cancel_search(self, *args):
        if self.filters or self.contexts:
            self.filters = []
            self.contexts = []
            self._status_entry_mode = 'search'
            self._activate_search(None)

    def _action_snooze(self, *args):
        self._start_entry('snooze', "Snooze: (t|p+|f)?[0-9]*[hdwmy]", needs_task=True)

    def _action_vi_mode(self, *args):
        self._start_entry('vicommand', "Vi emulation:")

    def _action_set_type(self, *args):
        self._start_entry('set_type', "Set type: (p+|f)? %s" % (",".join(Task.active_types)), needs_task=True)

    def _action_set_importance(self, *args):
        self._start_entry('set_importance', "Importance: (t|p+|f)?(0-%s)" % (cfg.max_importance), needs_task=True)

    def _action_set_lead_time(self, *args):
        self._start_entry('set_lead_time', "Lead time: (t|p+|f)?[0-9]*[hdwmy]", needs_task=True)

    def _action_edit_title(self, *args):
        title = ""
        t = self._get_current_task()
        if t: title = t.title
        self._start_entry('edit_title', "Title:", needs_task=True, text=title)

    def _action_set_due_date(self, *args):
        t = self._get_current_task()
        if not t:
            self.set_status("No task selected.")
            return

        # FIXME: this function is long and messy

        units = ('hours', 'days', 'weeks', 'months', 'years')
        unit_table = dict(h=0,d=1,w=2,m=3,y=4)

        g = self.gtk
        d = g.Dialog("Set Due Date", parent=self.window, flags=g.DIALOG_MODAL, 
                buttons=(g.STOCK_CANCEL, g.RESPONSE_REJECT, g.STOCK_OK, 
                    g.RESPONSE_ACCEPT))
        # configure year, month, day
        due_check = g.CheckButton("Due Date")
        cal = g.Calendar()
        d.vbox.pack_start(due_check, False, False)
        d.vbox.pack_start(cal, True, True)

        # hour and minute
        time_box = g.HBox(True, 5)
        d.vbox.pack_start(time_box, False, False)
        hour_spinner = g.SpinButton() #climb_rate=0.1, digits=0)
        minute_spinner = g.SpinButton() #climb_rate=0.1, digits=0)
        hour_spinner.set_increments(1,12)
        minute_spinner.set_increments(1,10)
        hour_spinner.set_range(0,23)
        minute_spinner.set_range(0,59)
        hour_spinner.set_numeric(True)
        minute_spinner.set_numeric(True)
        time_box.pack_start(g.Label("At:"), False, False)
        time_box.pack_start(g.Label("Hour"), False, False)
        time_box.pack_start(hour_spinner, False, False)
        time_box.pack_start(g.Label("Minute"), False, False)
        time_box.pack_start(minute_spinner, False, False)

        # allow configuring the recur schedule
        recur_box = g.HBox(False, 5)
        d.vbox.pack_start(recur_box, False, False)
        recur_check = g.CheckButton("Recurs")
        recur_box.pack_start(recur_check, False, False)
        recur_num_spinner = g.SpinButton()
        recur_num_spinner.set_increments(1,5)
        recur_num_spinner.set_range(1,99999)
        recur_num_spinner.set_numeric(True)
        recur_box.pack_start(g.Label(" every"), False, False)
        recur_box.pack_start(recur_num_spinner, False, False)
        recur_unit_box = g.combo_box_new_text()
        for unit in units:
            recur_unit_box.append_text(unit)
        recur_box.pack_start(recur_unit_box, False, False)

        if t.due_date: # fill fields with task's data
            date, interval = t.due_date[0], t.due_date[1:3]
            if date:
                due_check.set_active(True)
                cal.select_month(date[1]-1, date[0])
                cal.select_day(date[2])
                hour_spinner.set_value(date[3])
                minute_spinner.set_value(date[4])
            if interval[1]:
                recur_check.set_active(True)
                recur_num_spinner.set_value(interval[0])
                recur_unit_box.set_active(unit_table[interval[1]])

        # allow configuring the lead time
        lead_box = g.HBox(False, 5)
        d.vbox.pack_start(lead_box, False, False)
        lead_check = g.CheckButton("Lead time")
        lead_box.pack_start(lead_check, False, False)
        lead_num_spinner = g.SpinButton()
        lead_num_spinner.set_increments(1,5)
        lead_num_spinner.set_range(1,99999)
        lead_num_spinner.set_numeric(True)
        lead_box.pack_start(lead_num_spinner, False, False)
        lead_unit_box = g.combo_box_new_text()
        for unit in units:
            lead_unit_box.append_text(unit)
        lead_box.pack_start(lead_unit_box, False, False)

        if t.lead_time: # fill fields with task's data
            num, unit = t.lead_time
            lead_check.set_active(True)
            lead_num_spinner.set_value(num)
            lead_unit_box.set_active(unit_table[unit])

        d.show_all()

        response_id = d.run()

        d.hide()

        if response_id == g.RESPONSE_ACCEPT: # user clicked OK
            # handle the data entered, and save
            due = None
            if due_check.get_active(): # is due
                import time
                date = list(cal.get_date())
                date[1] += 1 # month is 0-based
                h = int(hour_spinner.get_value())
                m = int(minute_spinner.get_value())
                # build a time 9-tuple
                due = [date[0], date[1], date[2], h, m, 0]
                due.extend(time.localtime()[6:])
                due = tuple(due)

            recur = (None, None)
            if recur_check.get_active(): # recurs
                num = recur_num_spinner.get_value()
                unit = recur_unit_box.get_active_text()
                if unit:
                    unit = unit[:1]
                    recur = (num, unit)

            status = []

            # save to task
            if due or recur[1]:
                t.due_date = (due, recur[0], recur[1])
            else:
                if (t.due_date) and (t.due_date[0] or t.due_date[2]):
                    status.append("D=None")
                t.due_date = (None, None, None)

            lead = None
            if lead_check.get_active(): # has a lead time
                num = lead_num_spinner.get_value()
                unit = lead_unit_box.get_active_text()
                if unit:
                    unit = unit[:1]
                    lead = (num, unit)

            if t.lead_time and not lead:
                status.append("L=None")
            t.lead_time = lead

            # display status, and save
            date_text = ""
            if due: date_text = "%i-%02i-%02i@%02i:%02i" % due[:5]
            recur_text = ""
            if recur[0]: recur_text = "+%i%s" % recur
            if date_text or recur_text: status.append("D=%s%s" % (date_text, recur_text))

            if lead:
                status.append("L=%s%s" % lead)

            #self.set_status("Set D=%s%s" % (due_text, recur_text))
            self._save_task(t, "Set %s: %s" % (" ".join(status), t.full_title()))

    def _action_add_task(self, *args):
        self._start_entry('add_task', "New task title:", needs_task=True, text="")

    def _action_insert_task(self, *args):
        self._start_entry('insert_task', "New task title:", needs_task=True, text="")

    def _action_delete_task(self, *args):
        "Delete current task."
        t = self._get_current_task()
        title = t.full_title()
        self.set_status('Deleting task: %s' % (title))
        t.delete()
        self._save_task(t, 'Deleted task: %s' % (title))

    def _start_entry(self, mode, label, needs_task=False, text=''):
        "open an entry box, get some text, and pass it to a handler"
        # save the current task for later, in case the handler wants it
        t = self._get_current_task()
        self._status_entry_task = t
        if needs_task and not t:
            return

        # get user input
        self.set_status(label)
        self._status_entry_mode = mode
        self._status_entry.set_text(text)
        self._status_entry.show()
        self._status_entry.grab_focus()

    def _activate_status_entry(self, *args):
        self._status_entry.hide()
        text = self._status_entry.get_text()
        self._status_entry.set_text('')

        # find a handler, and call it
        mode = self._status_entry_mode
        try:
            func = getattr(self, "_activate_%s" % (mode))
        except:
            raise ValueError, "invalid mode %s" % (mode)

        try:
            func(text)
        except Exception, e:
            self.set_status(str(e))
            raise

        self._status_entry_mode = None
        self.treeview.grab_focus()

    def _activate_search(self, text):
        # filter visible data
        if text:
            self.filters.append(text)
            self.set_status("Search done.")
        else:
            self.filters = []
            self.set_status("Filter removed.")
        self._refresh()

    def _activate_context_filter(self, text):
        # filter visible data
        if text:
            self.contexts.append(text)
            self.set_status("Context filter done.")
        else:
            self.contexts = []
            self.set_status("Filter removed.")
        self._refresh()

    def _activate_snooze(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        t, text = self._get_rel(t, text)

        self.set_status("Snoozing %s" % (t.full_title()))
        try:
            t.snooze(text)
        except ValueError, e:
            self.set_status(str(e))
            return

        self._save_task(t, "Snoozed %s%s: %s" % (t.snoozed[1], t.snoozed[2], t.full_title()))

    def _activate_vicommand(self, text):
        """make a few vi commands work, like :w and :q and :e"""
        if text == "w":
            self._save_task_cb()
        elif text == "q":
            self._quit()
        #elif text == "wq": # FIXME: should save all files instead of just one?
        #    self._save_task_cb()
        #    self._quit()
        elif text == "e":
            self._action_reload_all()
        # this used to allow the user to prepend ":" to some TKDO commands,
        # but that seemed pointless, and added yet another place to hardcode
        # key bindings...  so it was removed
        else:
            self.set_status("Unknown vi command: %s" % (text))

    def _activate_set_type(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        if not text:
            self.set_status('Set type: aborted')
            return

        # allow setting parent or file type
        if text.startswith('p') or text.startswith('f'):
            t, text = self._get_rel(t, text)

        # TODO: allow setting task types to other states, like ...
        #   [S] skipped
        #   [F] failed
        # allow partial matches
        for choice in Task.active_types:
            if choice.startswith(text):
                text = choice
                break

        try:
            t.set_type(text)
        except ValueError, e:
            self.set_status(str(e))
            return

        self._save_task(t, "Set type (%s): %s" % (text, t.full_title()))

    def _activate_set_importance(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        t, text = self._get_rel(t, text)

        try:
            if text:
                imp = int(text)
            else:
                imp = None
            imp = t.set_importance(imp)
        except ValueError, e:
            self.set_status(str(e))
            return

        self._save_task(t, "Set I=%s: %s" % (imp, t.full_title()))

    def _activate_set_lead_time(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        t, text = self._get_rel(t, text)

        self.set_status("Setting lead time on %s" % (t.full_title()))
        try:
            # TODO: move this parsing to Task class
            import re
            pat = re.compile(r"^([.0-9]+)?([hdwmy])$")
            found = pat.search(text)
            if found:
                num, unit = found.group(1), found.group(2)
                if not num: num = 1
                t.lead_time = (num, unit)
            else:
                t.lead_time = None
        except Exception, e:
            self.set_status(str(e))
            return

        if t.lead_time:
            status = "L=%s%s" % (t.lead_time[0], t.lead_time[1])
        else:
            status = "L=None"
        self._save_task(t, "Set %s: %s" % (status, t.full_title()))

    def _activate_edit_title(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        if text:
            t.title = text
        else:
            self.set_status('Edit title: aborted.')
            return

        self._save_task(t, "Set new title: %s" % (t.full_title()))

    def _activate_add_task(self, text):
        return self._create_new_task(text, after=True)

    def _activate_insert_task(self, text):
        return self._create_new_task(text, after=False)

    def _create_new_task(self, text, after=False):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        if not text:
            self.set_status('New task: aborted.')
            return

        nt = t.add_sibling(text, after)

        #self.set_status("Created task: %s" % (nt.full_title()))
        self._save_task(nt, "Added task: %s" % (nt.full_title()))

    def _quit(self, *args):
        self.gtk.main_quit()

def counttabs(line):
    # FIXME: there has got to be a better way to do this
    i=0
    while i<len(line) and line[i] == '\t':
        i+=1
    return i

def fmt_date(when=None, fmt=None):
    import time
    if not fmt: fmt = datefmt
    if not when: when = time.localtime()
    if isinstance(when, float): when = time.localtime(when)
    text = time.strftime(fmt, when)
    return text

def parse_date(text):
    import time
    date = time.strptime(text, datefmt)
    #date = time.mktime(date)

    #print "parse_date(%s): %s" % (text, repr(date))
    return date

def date_add(date, num, unit):
    import time
    units = dict(h=(1,3), d=(1,2), w=(7,2), m=(1,1), y=(1,0))
    if unit not in units:
        raise ValueError, "bad unit: %s" % (unit)
    date = list(date)
    amt, field = units[unit]
    amt *= num
    date[field] += int(round(amt))
    date = tuple(date)
    # convert to epoch and back, to get a valid tuple
    date = time.mktime(date)
    date = time.localtime(date)

    return date

def date_diff(a, b):
    import time
    at = time.mktime(a)
    bt = time.mktime(b)
    diff = at - bt
    return diff

def duration(num, unit):
    """convert a readable time length into number of seconds.
    units are: h d w m y
    Use date_add instead, if you care about issues like leap years
    or daylight savings time.  This code gets those WRONG.
    """
    units = dict(
            h=60.0 * 60,
            d=60.0 * 60 * 24,
            w=60.0 * 60 * 24 * 7,
            m=60.0 * 60 * 24 * (365.0/12),
            y=60.0 * 60 * 24 * 365,
            )
    return num * units[unit]

def iround(num):
    "convert num (float) to str, and remove trailing zeroes"
    text = str(num)
    if text[-2:] == ".0":
        text = text[:-2]
    return text

def dedent(lines):
    """Remove leading whitespace from lines (same amount per line)."""
    first = lines[0]
    max_chars = len(first) - len(first.lstrip())
    if max_chars > 0:
        char = first[0]
        lines = [l.replace(char, '', max_chars) for l in lines]
    return lines

def show_diff(a,b, name="tmp"):
    import tempfile, os, sys
    f_a = tempfile.NamedTemporaryFile(prefix=name+'.')
    f_b = tempfile.NamedTemporaryFile(prefix=name+'.')
    f_a.write(a) ; f_a.flush()
    f_b.write(b) ; f_b.flush()
    sys.stdout.flush()
    err = os.system("diff -u %s %s" % (f_a.name, f_b.name))
    sys.stdout.flush()

def open_url(url, mode="rb"):
    """abstract file access to make URLs and remote files work"""
    # TODO: detect/save mtime (?)
    # (could be a pain for remote files)
    def expand_home(url):
        import os
        path = url
        path = path.replace("~", os.environ["HOME"])
        path = path.replace("$HOME", os.environ["HOME"])
        return path

    if "://" not in url:
        url = "file:///" + url
    # protocol:// (?), hostname (?), :port (?), path
    pat = re.compile(r"(\w+://)?([^/:]+)?(:\d+)?/(.+)")
    found = pat.search(url)
    if not found:
        raise ValueError, "invalid URL: %s" % repr(url)

    protocol = found.group(1)[:-3]
    hostname = found.group(2)
    portnum = found.group(3)
    if portnum: portnum = portnum[1:]
    path = found.group(4)

    if protocol == "file":
        path = expand_home(path)
        if ('w' in mode) and cfg.make_backups: # make backup first
            bak = path + '.bak'
            import shutil
            shutil.copy2(path, bak)
        fp = open(path, mode)
        return fp
    # TODO: add protocol for ssh, scp, sftp, or rsync
    else:
        raise ValueError, "unsupported protocol %s" % repr(protocol)

if __name__ == "__main__":
    import sys
    main(sys.argv[1:])
